import numpy as np
import matplotlib.pyplot as plt

class KNNregression:
    def __init__(self,
                 K:int=5,
                 W:str='Uniform',
                 p:int=2):
        self.K = K
        self.W = W.lower()
        self.p = p
    def Train(self,
              trX:np.ndarray,
              trY:np.ndarray):
        self.trX = trX.copy()
        self.trY = trY.copy()
    def Predict(self,
                X:np.ndarray) -> np.ndarray:
        n = X.shape[0]
        P = np.zeros((n, 1))
        for i, x in enumerate(X):
            dNs, yNs = self.GetNeighbors(x)
            w = self.GetWeights(dNs)
            P[i, 0] = np.multiply(yNs, w).sum()
        return P
    def GetNeighbors(self,
                     x:np.ndarray) -> tuple:
        d = x - self.trX
        Ds = np.linalg.norm(d, ord=self.p, axis=1)
        Is = np.argsort(Ds)
        iNs = Is[:self.K]
        dNs = Ds[iNs]
        yNs = self.trY[iNs, 0]
        return dNs, yNs
    def GetWeights(self,
                   dNs:np.ndarray) -> np.ndarray:
        if self.W == 'uniform':
            w = np.ones(self.K) / self.K
        elif self.W == 'distance':
            if dNs[0] > 1e-6:
                w = 1 / dNs
                w = w / w.sum()
            else:
                w = np.zeros(self.K)
                w[0] = 1
        return w
    def RegressionReport(self,
                         X:np.ndarray,
                         Y:np.ndarray,
                         Dataset:str):
        Range = Y.max() - Y.min()
        P = self.Predict(X)
        E = np.subtract(Y, P)
        MSE = np.power(E, 2).mean()
        RMSE = MSE ** 0.5
        NRMSE = 100 * RMSE / Range
        MAE = np.abs(E).mean()
        NMAE = 100 * MAE / Range
        MAPE = 100 * np.abs(E / Y).mean()
        R2 = 100 * (1 - MSE / np.var(Y))
        R = 10 * R2 ** 0.5
        print(f'KNN Regression Report For {Dataset} Dataset:')
        print(f'MSE:   {MSE:.4f}')
        print(f'RMSE:  {RMSE:.4f}')
        print(f'NRMSE: {NRMSE:.2f} %')
        print(f'MAE:   {MAE:.4f}')
        print(f'NMAE:  {NMAE:.2f} %')
        print(f'MAPE:  {MAPE:.2f} %')
        print(f'R2:    {R2:.2f} %')
        print(f'R:     {R:.2f} %')
        print('_' * 60)
    def RegressionPlot(self,
                       X:np.ndarray,
                       Y:np.ndarray,
                       Dataset:str):
        P = self.Predict(X)
        a = min(Y.min(), P.min())
        b = max(Y.max(), P.max())
        ab = np.array([a, b])
        plt.scatter(Y,
                    P,
                    s=20,
                    c='teal',
                    marker='s',
                    alpha=0.8,
                    label='Data')
        plt.plot(ab,
                 ab,
                 ls='-',
                 lw=1.2,
                 c='k',
                 label='Y=X')
        plt.plot(ab,
                 0.8 * ab,
                 ls='--',
                 lw=1,
                 c='r',
                 label='Y=0.8*X')
        plt.plot(ab,
                 1.2 * ab,
                 ls='--',
                 lw=1,
                 c='r',
                 label='Y=1.2*X')
        plt.title(f'KNN Regression Report For {Dataset} Dataset')
        plt.xlabel('Target Values')
        plt.ylabel('Predicted Values')
        plt.legend()
        plt.show()

np.random.seed(0)
plt.style.use('ggplot')

nD = 1000
nX = 3

sTr = 0.7

X = np.random.uniform(low=-1,
                      high=+1,
                      size=(nD, nX))

Y = 5 + 1.2*X[:, 0] - 2*X[:, 1]**2 + 4/(1+4*X[:, 2]**2)

Y = Y.reshape(-1, 1)

for i in range(nX):
    plt.scatter(X[:, i],
                Y[:, 0],
                s=20,
                c='teal',
                marker='o')
    plt.title('Data Scatter Plot')
    plt.xlabel(f'X{i + 1}')
    plt.ylabel('Y')
    plt.show()

nDtr = round(sTr * nD)

trX = X[:nDtr]
teX = X[nDtr:]

trY = Y[:nDtr]
teY = Y[nDtr:]

Model = KNNregression(W='distance')

Model.Train(trX, trY)

Model.RegressionReport(trX, trY, 'Train')
Model.RegressionReport(teX, teY, 'Test')

Model.RegressionPlot(trX, trY, 'Train')
Model.RegressionPlot(teX, teY, 'Test')
